#!/usr/bin/env python3
from eoepca_oidc import OpenIDClient

from client import Client

def main():

    # Create your client:
    # Parameters:
    resource_server_url = "my-RS-url.com"
    auth_server_url = "my-AS-url.com"
    OIDClient = OpenIDClient() # Configure OIDC with client creds!
    uma_cli = Client(resource_server_url, auth_server_url, OIDClient)

    uma_cli.request_resource("my/example/image.img",secure=False) # Use secure=True in production!


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from time import time

from eoepca_uma import rpt

def test_valid_token_intr_data():
    valid = [
        {"active": True, "permissions": [{"resource_id":"/simple/test", "resource_scopes": ["Auth"]}]},
        {"active": True, "permissions": [{"resource_id":"/simple/test/b", "resource_scopes": ["Auth", "Multiple", "Scopes"]}]},
    ]

    resources = [
        {"resource_id":"/simple/test",
         "resource_scopes": ["Auth"]},
        {"resource_id":"/simple/test/b",
         "resource_scopes": ["Auth","Multiple","Scopes"]},
        {"resource_id":"/simple/test/b",
         "resource_scopes": ["Auth","Scopes"]},
    ]

    for i in valid:
        assert(rpt.valid_token_introspection_data(i, resources=resources) == True)

def test_invalid_token_intr_data():
    invalid = [
        [],
        {},
        {"active": False },
        {"active": True, "permissions": [{"resource_id":"/simple/test", "resource_scopes": ["AAAAA"]}]},
        {"active": True, "permissions": [{"resource_id":"/simple/invalid", "resource_scopes": ["Auth"]}]},
        {"active": True, "permissions": [{"resource_id":"/simple/invalid", "resource_scopes": ["BBBB"]}]},
        {"active": True, "permissions": [{"resource_id":"/simple/test/b", "resource_scopes": ["Auth", "Multiple","Scopes"]}]},
        {"active": True, "permissions": [{"resource_id":"/simple/test/b", "resource_scopes": []}]},
        {"active": True, "permissions": [{"resource_id":"/simple/test/b", "resource_scopes": ["Auth"]}]},
    ]

    resources = [
        {"resource_id":"/simple/test",
         "resource_scopes": ["Auth"]},
        {"resource_id":"/simple/test/b",
         "resource_scopes": ["Auth","Multiple","Scopes","Invalid"]},
    ]


    for i in invalid:
        assert(rpt.valid_token_introspection_data(i,resources=resources) == False)


def test_time_valid_token_intr_data():
    now = time()

    valid = [
        {"exp" : now+1_000_000},
        {"exp" : now+100},
        {"exp" : now+5},

        {"nbf" : now-5},
        {"nbf" : now-1000},

        {"iat" : now-5},
        {"iat" : now-1000},

        # Combination
        {
            "exp" : now+1000,
            "nbf" : now-100,
            "iat" : now-1000
        }
    ]

    resources = [
        {"resource_id":"/simple/test",
         "resource_scopes": ["Auth"]},
    ]

    for i in valid:
        # We are just testing time here
        i["active"] = True
        i["permissions"]= [{"resource_id":"/simple/test", "resource_scopes": ["Auth"]}]
        assert(rpt.valid_token_introspection_data(i, resources=resources) == True)


def test_time_invalid_token_intr_data():
    now = time()

    valid = [
        {"exp" : now-1_000_000},
        {"exp" : now-100},
        {"exp" : now-5},
        {"exp" : now},

        {"nbf" : now+5},
        {"nbf" : now+1000},

        {"iat" : now+5},
        {"iat" : now+1000},

        # Combination
        {
            "exp" : now-1000,
            "nbf" : now+100,
            "iat" : now+1000
        }
    ]

    resources = [
        {"resource_id":"/simple/test",
         "resource_scopes": ["Auth"]},
    ]

    for i in valid:
        # Check time validity, even with an active true
        i["active"] = True
        i["permissions"]= [{"resource_id":"/simple/test", "resource_scopes": ["Auth"]}]
        assert(rpt.valid_token_introspection_data(i, resources=resources) == False)



#!/usr/bin/env python3
from typing import Any
from urllib3 import disable_warnings
from urllib3.exceptions import InsecureRequestWarning

from requests import Response

def dict_insert_if_exists(d: dict, k: Any, v: Any):
    """
    Internal utility which adds a key and a value to a dictionary,
    given that the key and the value are not None or empty.

    Does nothing if any of the values is invalid.

    EDITS THE DICTIONARY BY REFERENCE
    """

    if k is None or k is "":
        return
    if v is None or v is "" or len(v) is 0:
        return
    
    d[k] = v

def disable_warnings_if_debug(secure: bool):
    """
    Deactivates warnings from Requests when not checking SSL certs, useful for debugging
    """
    if not secure:
        disable_warnings(InsecureRequestWarning)

def is_ok(response: Response):
    """
    Returns True or False depending on the response code of the server 
    """
    return response.status_code > 199 and response.status_code < 230 and "error" not in response.text.lower()



#!/usr/bin/env python3
from requests import request, Response
from WellKnownHandler import WellKnownHandler, TYPE_OIDC, KEY_OIDC_TOKEN_ENDPOINT
from WellKnownHandler import TYPE_UMA_V2, KEY_UMA_V2_TOKEN_ENDPOINT
from eoepca_oidc import OpenIDClient

from eoepca_uma.utils import is_ok
from eoepca_uma.rpt import request_for_rpt

class Client:
    """
    UMA Client implementation,
    acting on behalf of the requesting party.

    Example:
    c = Client(<parameters>)
    img = c.request_resource("/my/image.jpg",<...>)
    """

    def __init__(self, resource_server_url: str, auth_server_url: str, OIDClient: OpenIDClient, secure: bool = True):
        """
        Creates a new UMA Client instance.

        - CAN THROW EXCEPTIONS
        - MAKES A CONNECTION TO AN EXTERNAL ENDPOINT

        Args:
        - resource_server_url = URL of the resource server
        - auth_server_url = URL of the Authorization Server
        - OIDClient = Instance of a OpenIDClient (from eoepca_oidc), configured for the same Auth server.
        - secure (Optional) = toggle checking of SSL certificates. Activating this is recommended on production environments
        
        Returns:
            A configured UMA Client, or an exception
        """
        self.resource_server = resource_server_url

        self.wkh = WellKnownHandler(auth_server_url, secure=secure)
        self.oidc_client = OIDClient

    def request_resource(self, uri: str, rpt: str = None, secure: bool = True) -> bytes:
        """
        Requests a resource from the resource server.

        If you have an RPT token for this resource, you can try to re-use it calling
        this function with your RPT, but please notice this function wont try to re-generate
        an RPT via ticket if one is given.

        If you have no RPT but you have initialized this client with the proper credentials,
        this function call will auto-authenticate, handling ticket and RPT calls until the
        resource is obtained.

        - CAN THROW EXCEPTIONS
        - MAKES A CONNECTION TO AN EXTERNAL ENDPOINT

        Args:
        - uri = URI to the resource you want to access inside the resource server.
        - rpt (Optional) = String containing the rpt (token)
        - secure (Optional) = toggle checking of SSL certificates. Activating this is recommended on production environments
        
        Returns:
            The resource requested in bytes, or an exception if something went wrong.
        """
        headers = {}
        if rpt:
            headers = {"Authorization": "Bearer "+rpt}

        # Request resource
        ret = request("GET", self.resource_server + uri, headers=headers, secure=secure)

        # Handle ticket
        if ret.status_code == 401:
            rpt = self._handle_ticket_request(ret)
            # Re-try with an rpt obtained from ticket
            return self.request_resource(uri, rpt, secure)

        # Any error other than a 401 is an error that this client cannot automatically solve
        elif not is_ok(ret):
            raise Exception("Resource server denied access with an unexpected error: "+str(ret.status_code)+": "+str(ret.reason))
                
        # Return resource when access is achieved
        return ret.content

    def _handle_ticket_request(self, response: Response, secure: bool = True) -> str:
        """
        Handles the ticket answer when attempting to access to a resource.

        - CAN THROW EXCEPTIONS
        - MAKES A CONNECTION TO AN EXTERNAL ENDPOINT

        Args:
        - response = Response object from requests package, after requesting access to a resource.
        - secure (Optional) = toggle checking of SSL certificates. Activating this is recommended on production environments
        
        Returns:
            RPT for this resource, or an exception
        """
        if "WWW-Authenticate" not in response.headers:
            raise Exception("Response from server does not include necessary header: 'WWW-Authenticate'")

        data = response.headers["WWW-Authenticate"]
        # Get ticket
        data = data.strip().split(",")
        ticket= [x.split("=")[1].strip('"') for x in data if "ticket" in x ]

        # Get OIDC token
        token_endpoint = self.wkh.get(TYPE_OIDC , KEY_OIDC_TOKEN_ENDPOINT)
        self.oidc_client.postRequestToken({"token_endpoint": token_endpoint}, verify=secure)
        client_creds_token= self.oidc_client.token

        token_endpoint = self.wkh.get(TYPE_UMA_V2, KEY_UMA_V2_TOKEN_ENDPOINT)
        rpt_res = request_for_rpt(client_creds_token, token_endpoint, ticket, secure=secure)
        
        return rpt_res["access_token"]

    


#!/usr/bin/env python3
from requests import request
from typing import List

from eoepca_uma.utils import dict_insert_if_exists
from eoepca_uma.utils import disable_warnings_if_debug
from eoepca_uma.utils import is_ok

# Client utilities

def read(pat: str, resource_registration_endpoint: str,
         resource_id: str,
         secure: bool = False) -> dict:
    """
    Reads the information for a single resource, indicated by resource ID.

    Note that if an empty resource ID is given, this function is identical to 'list()', but please use each function
    for their respective intended use as this behaviour is not intended by design and may change at any time

    - CAN THROW EXCEPTIONS
    - MAKES A CONNECTION TO AN EXTERNAL ENDPOINT

    Args:
    - pat = String containing the pat (token)
    - resource_registration_endpoint = URL of the resource registration endpoint in the AS
    - resource_id = ID of the resource on the AS
    - secure = toggle checking of SSL certificates. Activating this is recommended on production environments
    
    Returns:
        JSON-formatted information about the resource
    """
    headers={"Authorization": "Bearer "+pat}

    disable_warnings_if_debug(secure)
    if resource_registration_endpoint[-1] is not "/":
        resource_registration_endpoint += "/"

    response = request("GET", resource_registration_endpoint + resource_id, headers=headers, verify=secure)
    if not is_ok(response):
        raise Exception("An error occurred while getting a resource's information: "+str(response.status_code)+":"+str(response.reason)+":"+str(response.text))

    return response.json()

def list(pat: str, resource_registration_endpoint: str,
         secure: bool = False) -> List[str]:
    """
    Lists all previously registered resources by ID for this resource owner.
    - CAN THROW EXCEPTIONS
    - MAKES A CONNECTION TO AN EXTERNAL ENDPOINT

    Args:
    - pat = String containing the pat (token)
    - resource_registration_endpoint = URL of the resource registration endpoint in the AS
    - secure = toggle checking of SSL certificates. Activating this is recommended on production environments
    
    Returns:
        List of IDs (str) registered
    """
    headers={"Authorization": "Bearer "+pat}

    disable_warnings_if_debug(secure)
    response = request("GET", resource_registration_endpoint, headers=headers, verify=secure)

    if not is_ok(response):
        raise Exception("An error occurred while listing resources: "+str(response.status_code)+":"+str(response.reason)+":"+str(response.text))

    return response.json()

def create(pat: str, resource_registration_endpoint: str,
             name: str, scopes: List[str],
             description: str = None, icon_uri: str = None, typ: str = None,
             secure: bool = False) -> str:
    """
    Registers a new resource in the AS.
    - CAN THROW EXCEPTIONS
    - MAKES A CONNECTION TO AN EXTERNAL ENDPOINT

    Args:
    - pat = String containing the pat (token)
    - resource_registration_endpoint = URL of the resource registration endpoint in the AS
    - name = Name given to the new resource
    - scopes = List of scopes (strings) assigned to this resource
    - description (Optional) = Description for the resource
    - icon_uri (Optional) = URI to an icon representing this resource
    - typ (Optional) = Type (string/URI) of this resource
    - secure = toggle checking of SSL certificates. Activating this is recommended on production environments
    
    Returns:
        Resource ID given by the AS associated with the newly created resource, or an error
    """

    payload = {"name": name , "resource_scopes": scopes }
    dict_insert_if_exists(payload, "description", description)
    dict_insert_if_exists(payload, "icon_uri", icon_uri)
    dict_insert_if_exists(payload, "type", typ)

    headers = {
        'Content-Type': "application/json",
        'Authorization': "Bearer "+pat,
    }

    disable_warnings_if_debug(secure)
    response = request("POST", resource_registration_endpoint, json=payload, headers=headers, verify=secure)

    if not is_ok(response):
        raise Exception("An error occurred while registering the resource: "+str(response.status_code)+":"+str(response.reason)+":"+str(response.text))


    try:
        return response.json()["_id"]
    except Exception as e:
        raise Exception("Call to registration endpoint returned unexpected value: '"+response.text+"'"+". Error: "+str(e))

def delete(pat: str, resource_registration_endpoint: str,
           resource_id: str,
           secure: bool = False):
    """
    Deletes a resource from the AS.
    - CAN THROW EXCEPTIONS
    - MAKES A CONNECTION TO AN EXTERNAL ENDPOINT

    Args:
    - pat = String containing the pat (token)
    - resource_registration_endpoint = URL of the resource registration endpoint in the AS
    - resource_id = ID of the resource
    - secure = toggle checking of SSL certificates. Activating this is recommended on production environments
    
    Returns:
        Nothing. If no exceptions are raised, the operation completed succesfully.
    """
    if resource_registration_endpoint[-1] is not "/":
        resource_registration_endpoint += "/"

    headers={"Authorization": "Bearer "+pat}

    disable_warnings_if_debug(secure)
    response = request("DELETE", resource_registration_endpoint + resource_id, headers=headers, verify=secure)

    if not is_ok(response):
        raise Exception("An error occurred while deleting the resource: "+str(response.status_code)+":"+str(response.reason)+":"+str(response.text))

def update(pat: str, resource_registration_endpoint: str,
            resource_id: str,
            name: str, scopes: List[str],
            description: str = None, icon_uri: str = None, typ: str = None,
            secure: bool = False) -> str:
    """
    Updates one or more aspects of the resource indicated by id.
    
    The entirety of the resource will be overwritten with these values as per UMA 2.0's standard.
    https://docs.kantarainitiative.org/uma/wg/rec-oauth-uma-federated-authz-2.0.html#update-resource-set
 
    - CAN THROW EXCEPTIONS
    - MAKES A CONNECTION TO AN EXTERNAL ENDPOINT

    Args:
    - pat = String containing the pat (token)
    - resource_registration_endpoint = URL of the resource registration endpoint in the AS
    - resource_id = ID of the resource
    - name (Optional) = Name given to the new resource
    - scopes (Optional) = List of scopes (strings) assigned to this resource
    - description (Optional) = Description for the resource
    - icon_uri (Optional) = URI to an icon representing this resource
    - typ (Optional) = Type (string/URI) of this resource
    - secure = toggle checking of SSL certificates. Activating this is recommended on production environments
    
    Returns:
        Resource ID given by the AS associated with the edited resource
    """
    payload = {}
    dict_insert_if_exists(payload, "name", name)
    dict_insert_if_exists(payload, "resource_scopes", scopes)
    dict_insert_if_exists(payload, "description", description)
    dict_insert_if_exists(payload, "icon_uri", icon_uri)
    dict_insert_if_exists(payload, "type", typ)

    if len(payload) == 0:
        raise Exception("No attribute to update the resource with, payload empty")

    if resource_registration_endpoint[-1] is not "/":
        resource_registration_endpoint += "/"

    headers = {
        'content-type': "application/json",
        'authorization': "Bearer "+pat,
    }

    disable_warnings_if_debug(secure)
    response = request("PUT", resource_registration_endpoint + resource_id, json=payload, headers=headers, verify=secure)

    if not is_ok(response):
        raise Exception("An error occurred while registering the resource: "+str(response.status_code)+":"+str(response.reason)+":"+str(response.text))

    try:
        return response.json()["_id"]
    except Exception as e:
        raise Exception("Call to registration endpoint returned unexpected value: '"+response.text+"'"+". Error: "+str(e))

def delete_all(pat: str, resource_registration_endpoint: str, secure: bool = False):
    """
    Deletes ALL resources that this resource owner owns from the AS.

    Note that the deletion can stop mid-process in case of an error, and thus only deleting part
    or none of the resources.

    - CAN THROW EXCEPTIONS
    - MAKES A CONNECTION TO AN EXTERNAL ENDPOINT

    Args:
    - pat = String containing the pat (token)
    - resource_registration_endpoint = URL of the resource registration endpoint in the AS
    - secure = toggle checking of SSL certificates. Activating this is recommended on production environments
    
    Returns:
        Nothing. If no exceptions are raised, the operation completed succesfully.
    """
    
    all_resources = list(pat,resource_registration_endpoint,secure)

    for resource_id in all_resources:
        delete(pat, resource_registration_endpoint, resource_id, secure)

# Resource Server utilities

def request_access_ticket(pat: str, permission_endpoint: str,
                        resources: List[dict],
                        secure: bool = False) -> str:
    """
    As a Resource Server, request permission to the AS to access a resource,
    generating a ticket as a result.

    - CAN THROW EXCEPTIONS
    - MAKES A CONNECTION TO AN EXTERNAL ENDPOINT

    Args:
    - pat = String containing the pat (token)
    - permission_endpoint = URL of the token permission endpoint in the AS 
    - resources = List of resources to request permission to.
                            Format:
                            [
                                {
                                    "resource_id": <str resource id>,
                                    "resource_scopes": [ <scope 1>, <scope 2>, ...]
                                },
                                ...
                            ]
    - secure = toggle checking of SSL certificates. Activating this is recommended on production environments
    
    Returns:
        A string containing the ticket for accessing those resources.
    """

    headers = {
        'content-type': "application/json",
        'authorization': "Bearer "+pat,
    }

    if len(resources) == 1:
        resources = resources[0] # Use a single dict instead of a list for 1 resource

    disable_warnings_if_debug(secure)
    response = request("POST", permission_endpoint , json=resources, headers=headers, verify=secure)

    if not is_ok(response):
        raise Exception("An error occurred while requesting permission for a resource: "+str(response.status_code)+":"+str(response.reason)+":"+str(response.text))

    try:
        return response.json()["ticket"]
    except Exception as e:
        raise Exception("Call to permission endpoint returned unexpected value or error: '"+response.text+"'"+". Error: "+str(e))



#!/usr/bin/env python3
from requests import request
from time import time
from typing import List

from eoepca_uma.utils import disable_warnings_if_debug, is_ok

def introspect(rpt: str, pat: str, introspection_endpoint: str, secure: bool = False) -> dict:
    """
    Gets information about an RPT, using the AS' introspection endpoint.
    - CAN THROW EXCEPTIONS
    - MAKES A CONNECTION TO AN EXTERNAL ENDPOINT

    Args:
    - rpt = String containing the rpt (token)
    - pat = String containing the pat (token)
    - introspection_endpoint = String containing the url to the AS' introspection endpoint
    - secure = toggle checking of SSL certificates. Activating this is recommended on production environments
    
    Returns:
        JSON-formatted data about the RPT, or an error from the AS
    """

    headers = {
        'content-type': "application/x-www-form-urlencoded",
        'authorization': "Bearer "+pat,
        }

    payload = "token="+rpt

    disable_warnings_if_debug(secure)
    r = request("POST", introspection_endpoint, headers=headers, data=payload, verify=secure)

    if not is_ok(r):
        raise Exception("An error occurred while registering the resource: "+str(r.status_code)+":"+str(r.reason))

    try:
        return r.json()
    except Exception as e:
        raise Exception("Call to introspection point returned unexpected value: "+str(e))

def valid_token_introspection_data(rpt_info: dict, resources: List[dict], time_margin: float = 0.0) -> bool:
    """
    Asserts if a given rpt is valid at this time, using the introspection data from it, in regards to the resources and scopes requested.
    Keep in mind this will return False if any scope or resource is denied.
    
    This validation is done in compliance with RFC7662 'OAuth 2.0 Token Introspection'.

    Keep in mind that if the data analyzed is outdated, this function may return the wrong answer, since it doesn't
    update with the AS.
    More precisely, this function WILL detect when a token is outdated or otherwise not usable at this specific date,
    but NOT when scopes have changed, or any other remote action the AS might have taken or changed.

    It is crucial for the time-checking to work that this code is executed on a time-synced server, otherwise the
    time meassurements could be off.

    Time check uses the following fields: exp, iat, nbf

    Alternatively, you can use 'is_valid_now' to get updated RPT information and check it in a single call. 

    Args:
    - rpt_info = Dict containing the data from an RPT, obtained from a correct 'introspect' call
    - resources = List of resources to assert permission over.
                            Format:
                            [
                                {
                                    "resource_id": <str resource id>,
                                    "resource_scopes": [ <scope 1>, <scope 2>, ...]
                                },
                                ...
                            ]
    - time_margin (Optional) = A time difference to add to the current timestamp, to ensure this token will be valid _at least until_ now + time_margin
   
    Note that there might not be time information in the data, and thus 'time_margin' would do nothing.


    Returns:
        True if the token is valid at this time given this information, False otherwise.
    """
    now = time()

    if "exp" in rpt_info and rpt_info["exp"] is not None:
        if rpt_info["exp"] < (now - time_margin) : return False

    # Not a time validity check per se, but a sanity check for corrupt / invalid data nontheless
    if "iat" in rpt_info and rpt_info["iat"] is not None:
        if rpt_info["iat"] > (now + time_margin) : return False
    
    if "nbf" in rpt_info and rpt_info["nbf"] is not None:
        if rpt_info["nbf"] > (now + time_margin) : return False
    
    # Actual check of validity using the AS' criteria
    if "active" not in rpt_info or rpt_info["active"] != True: return False

    # Check against the resources:
    if "permissions" not in rpt_info: return False

    for requested_resoure in resources:
        requested_r = requested_resoure["resource_id"]
        requested_scopes = requested_resoure["resource_scopes"]
        ok_r = False
        for allowed_resource in rpt_info["permissions"]:
            allowed_r = allowed_resource["resource_id"]
            allowed_scopes = allowed_resource["resource_scopes"]
            if requested_r == allowed_r:
                ok_r = True
                for scope in requested_scopes:
                    if scope not in allowed_scopes:
                        return False
                break
        
        if ok_r: return True

    # Fallthrough to false, as a secure default
    return False


def is_valid_now(rpt: str, pat: str, introspection_endpoint: str, resources: List[dict], time_margin: float = 0.0, secure: bool = False) -> bool:
    """
    Uses the introspection from the AS to get updated data about the RPT, then returns True/False
    if the RPT is valid or not at the moment of the call of this function, in regards to the resources and scopes requested.
    Keep in mind this will return False if any scope or resource is denied.

    This is equal to calling 'introspect' and then 'valid_token_introspection_data'.

    - CAN THROW EXCEPTIONS
    - MAKES A CONNECTION TO AN EXTERNAL ENDPOINT

    Args:
    - rpt = String containing the rpt (token)
    - pat = String containing the pat (token)
    - introspection_endpoint = String containing the url to the AS' introspection endpoint
    - resources = List of resources to assert permission over.
                            Format:
                            [
                                {
                                    "resource_id": <str resource id>,
                                    "resource_scopes": [ <scope 1>, <scope 2>, ...]
                                },
                                ...
                            ]
    - time_margin (Optional) = A time difference to add to the current timestamp, to ensure this token will be valid _at least until_ now + time_margin
    - secure = toggle checking of SSL certificates. Activating this is recommended on production environments
    Note that there might not be time information in the data, and thus 'time_margin' would do nothing.
    
    Returns:
        True if the token is valid at this time given this information, False otherwise.
    """

    try:
        introspection_data = introspect(rpt, pat, introspection_endpoint, secure)
    except:
        return False
    
    return valid_token_introspection_data(introspection_data, resources, time_margin)


def request_for_rpt(client_creds_token: str,
                    token_endpoint: str, ticket: str,
                    claim_token: str = None, claim_token_format: str = None,
                    scopes: List[str] = None,
                    rpt: str = None,
                    pct: str = None,
                    secure: bool = True) -> dict:
    """
    Requests the AS for an RPT given the arguments used.

    Reference: https://docs.kantarainitiative.org/uma/wg/rec-oauth-uma-grant-2.0.html#uma-grant-type

    - CAN THROW EXCEPTIONS
    - MAKES A CONNECTION TO AN EXTERNAL ENDPOINT

    Args:
    - client_creds_token = token obtained from AS with client credentials
    - token_endpoint = String containing the url to the AS' token endpoint
    - ticket = Ticket obtained from Resource Server for this request
    - claim_token (Optional) = A string containing directly pushed claim information in the indicated format. It MUST be base64url encoded unless specified otherwise by the claim token format
    - claim_token_format (Optional) = If this parameter is used, it MUST appear together with the claim_token parameter. A string specifying the format of the claim token in which the client is directly pushing claims to the authorization server. The string MAY be a URI
    - scopes (Optional) = List of requested scopes
    - rpt (Optional) = A string containing an existing RPT. Allows the Auth Server to upgrade instead of issuing a new RPT.
    - pct (Optional) = A string with an old PCT to optimize looking for a new RPT.
    - secure = toggle checking of SSL certificates. Activating this is recommended on production environments
    
    Returns:
        A dictionary with at least an access_token and a 'token_type', or an exception
    """
    grant_type = "urn:ietf:params:oauth:grant-type:uma-ticket" # Fixed value according to https://docs.kantarainitiative.org/uma/wg/rec-oauth-uma-grant-2.0.html#uma-grant-type

    headers = {"Authorization": "Basic "+ client_creds_token}

    payload = "grant_type="+grant_type+"&ticket="+ticket
    if claim_token:
        payload+="&claim_token="+claim_token
        # Format only allowed when claim token is given.
        if claim_token_format:
            payload+="&claim_token_format="+claim_token_format
    
    if scopes and len(scopes) > 0:
        payload+="&scope="+" ".join(scopes)
    
    if rpt and len(rpt) > 0:
        payload+="&rpt="+rpt
    if pct and len(pct) > 0:
        payload+="&pct="+pct

    disable_warnings_if_debug(secure)
    response = request("POST", token_endpoint , data=payload, headers=headers, verify=secure)

    if not is_ok(response):
        raise Exception("An error occurred while requesting RPT: "+str(response.status_code)+":"+str(response.reason))

    return response.json()

